#  Copyright 2024 Hathor Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from hathorlib.headers import VertexBaseHeader, VertexHeaderId


class VertexParser:
    __slots__ = ()

    @staticmethod
    def get_supported_headers() -> dict[VertexHeaderId, type[VertexBaseHeader]]:
        """Return a dict of supported headers."""
        from hathorlib.headers import NanoHeader, VertexHeaderId
        return {
            VertexHeaderId.NANO_HEADER: NanoHeader,
        }

    @staticmethod
    def get_header_parser(header_id_bytes: bytes) -> type[VertexBaseHeader]:
        """Get the parser for a given header type."""
        from hathorlib.headers import VertexHeaderId
        header_id = VertexHeaderId(header_id_bytes)
        supported_headers = VertexParser.get_supported_headers()
        if header_id not in supported_headers:
            raise ValueError(f'Header type not supported: {header_id_bytes!r}')
        return supported_headers[header_id]



# Copyright 2023 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from typing import TYPE_CHECKING

from hathorlib.headers.base import VertexBaseHeader
from hathorlib.headers.types import VertexHeaderId
from hathorlib.nanocontracts import DeprecatedNanoContract
from hathorlib.nanocontracts.types import NCActionType
from hathorlib.utils import decode_unsigned, encode_unsigned, int_to_bytes, unpack, unpack_len

if TYPE_CHECKING:
    from hathorlib.base_transaction import BaseTransaction

NC_INITIALIZE_METHOD = 'initialize'
ADDRESS_LEN_BYTES = 25
ADDRESS_SEQNUM_SIZE: int = 8  # bytes
_NC_SCRIPT_LEN_MAX_BYTES: int = 2


@dataclass(frozen=True)
class NanoHeaderAction:
    type: NCActionType
    token_index: int
    amount: int


@dataclass(frozen=True)
class NanoHeader(VertexBaseHeader):
    tx: BaseTransaction

    # Sequence number for the caller.
    nc_seqnum: int

    # nc_id equals to the blueprint_id when a Nano Contract is being created.
    # nc_id equals to the nanocontract_id when a method is being called.
    nc_id: bytes

    # Name of the method to be called. When creating a new Nano Contract, it must be equal to 'initialize'.
    nc_method: str

    # Serialized arguments to nc_method.
    nc_args_bytes: bytes

    nc_actions: list[NanoHeaderAction]

    # Address and script with signature(s) of the transaction owner(s)/caller(s). Supports P2PKH and P2SH.
    nc_address: bytes
    nc_script: bytes

    @classmethod
    def _deserialize_action(cls, buf: bytes) -> tuple[NanoHeaderAction, bytes]:
        from hathorlib.base_transaction import bytes_to_output_value
        type_bytes, buf = buf[:1], buf[1:]
        action_type = NCActionType.from_bytes(type_bytes)
        (token_index,), buf = unpack('!B', buf)
        amount, buf = bytes_to_output_value(buf)
        return NanoHeaderAction(
            type=action_type,
            token_index=token_index,
            amount=amount,
        ), buf

    @classmethod
    def deserialize(cls, tx: BaseTransaction, buf: bytes) -> tuple[NanoHeader, bytes]:
        header_id, buf = buf[:1], buf[1:]
        assert header_id == VertexHeaderId.NANO_HEADER.value

        nc_id, buf = unpack_len(32, buf)
        nc_seqnum, buf = decode_unsigned(buf, max_bytes=ADDRESS_SEQNUM_SIZE)
        (nc_method_len,), buf = unpack('!B', buf)
        nc_method, buf = unpack_len(nc_method_len, buf)
        (nc_args_bytes_len,), buf = unpack('!H', buf)
        nc_args_bytes, buf = unpack_len(nc_args_bytes_len, buf)

        nc_actions: list[NanoHeaderAction] = []
        if not isinstance(tx, DeprecatedNanoContract):
            (nc_actions_len,), buf = unpack('!B', buf)
            for _ in range(nc_actions_len):
                action, buf = cls._deserialize_action(buf)
                nc_actions.append(action)

        nc_address, buf = unpack_len(ADDRESS_LEN_BYTES, buf)
        nc_script_len, buf = decode_unsigned(buf, max_bytes=_NC_SCRIPT_LEN_MAX_BYTES)
        nc_script, buf = unpack_len(nc_script_len, buf)

        decoded_nc_method = nc_method.decode('ascii')

        return cls(
            tx=tx,
            nc_seqnum=nc_seqnum,
            nc_id=nc_id,
            nc_method=decoded_nc_method,
            nc_args_bytes=nc_args_bytes,
            nc_actions=nc_actions,
            nc_address=nc_address,
            nc_script=nc_script,
        ), bytes(buf)

    @staticmethod
    def _serialize_action(action: NanoHeaderAction) -> bytes:
        from hathorlib.base_transaction import output_value_to_bytes
        ret = [
            action.type.to_bytes(),
            int_to_bytes(action.token_index, 1),
            output_value_to_bytes(action.amount),
        ]
        return b''.join(ret)

    def _serialize_without_header_id(self, *, skip_signature: bool) -> deque[bytes]:
        """Serialize the header with the option to skip the signature."""
        encoded_method = self.nc_method.encode('ascii')

        ret: deque[bytes] = deque()
        ret.append(self.nc_id)
        ret.append(encode_unsigned(self.nc_seqnum, max_bytes=ADDRESS_SEQNUM_SIZE))
        ret.append(int_to_bytes(len(encoded_method), 1))
        ret.append(encoded_method)
        ret.append(int_to_bytes(len(self.nc_args_bytes), 2))
        ret.append(self.nc_args_bytes)

        if not isinstance(self.tx, DeprecatedNanoContract):
            ret.append(int_to_bytes(len(self.nc_actions), 1))
            for action in self.nc_actions:
                ret.append(self._serialize_action(action))

        ret.append(self.nc_address)
        if not skip_signature:
            ret.append(encode_unsigned(len(self.nc_script), max_bytes=_NC_SCRIPT_LEN_MAX_BYTES))
            ret.append(self.nc_script)
        else:
            ret.append(encode_unsigned(0, max_bytes=_NC_SCRIPT_LEN_MAX_BYTES))
        return ret

    def serialize(self) -> bytes:
        ret = self._serialize_without_header_id(skip_signature=False)
        ret.appendleft(VertexHeaderId.NANO_HEADER.value)
        return b''.join(ret)

    def get_sighash_bytes(self) -> bytes:
        ret = self._serialize_without_header_id(skip_signature=True)
        return b''.join(ret)



# Copyright 2023 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from enum import Enum, unique

from hathorlib.utils import bytes_to_int, int_to_bytes


@unique
class NCActionType(Enum):
    """Types of interactions a transaction might have with a contract."""
    DEPOSIT = 1
    WITHDRAWAL = 2
    GRANT_AUTHORITY = 3
    ACQUIRE_AUTHORITY = 4

    def __str__(self) -> str:
        return self.name.lower()

    def to_bytes(self) -> bytes:
        return int_to_bytes(number=self.value, size=1)

    @staticmethod
    def from_bytes(data: bytes) -> NCActionType:
        return NCActionType(bytes_to_int(data))



# Copyright 2025 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import struct
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Iterator, TypeVar, overload

from typing_extensions import Self

from .types import Buffer

if TYPE_CHECKING:
    from .adapters import MaxBytesDeserializer
    from .bytes_deserializer import BytesDeserializer

T = TypeVar('T')


class Deserializer(ABC):
    def finalize(self) -> None:
        """Check that all bytes were consumed, the deserializer cannot be used after this."""
        raise TypeError('this deserializer does not support finalization')

    @staticmethod
    def build_bytes_deserializer(data: Buffer) -> BytesDeserializer:
        from .bytes_deserializer import BytesDeserializer
        return BytesDeserializer(data)

    @abstractmethod
    def is_empty(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def peek_byte(self) -> int:
        """Read a single byte but don't consume from buffer."""
        raise NotImplementedError

    @abstractmethod
    def peek_bytes(self, n: int, *, exact: bool = True) -> Buffer:
        """Read n single byte but don't consume from buffer."""
        raise NotImplementedError

    def peek_struct(self, format: str) -> tuple[Any, ...]:
        size = struct.calcsize(format)
        data = self.peek_bytes(size)
        return struct.unpack(format, data)

    @abstractmethod
    def read_byte(self) -> int:
        """Read a single byte as unsigned int."""
        raise NotImplementedError

    @abstractmethod
    def read_bytes(self, n: int, *, exact: bool = True) -> Buffer:
        """Read n bytes, when exact=True it errors if there isn't enough data"""
        # XXX: this is a blanket implementation that is an example of the behavior, this implementation has to be
        #      explicitly used if needed
        def iter_bytes() -> Iterator[int]:
            for _ in range(n):
                if not exact and self.is_empty():
                    break
                yield self.read_byte()
        return bytes(iter_bytes())

    @abstractmethod
    def read_all(self) -> Buffer:
        """Read all bytes until the reader is empty."""
        # XXX: it is recommended that implementors of Deserializer specialize this implementation
        def iter_bytes() -> Iterator[int]:
            while not self.is_empty():
                yield self.read_byte()
        return bytes(iter_bytes())

    def read_struct(self, format: str) -> tuple[Any, ...]:
        size = struct.calcsize(format)
        data = self.read_bytes(size)
        return struct.unpack_from(format, data)

    def with_max_bytes(self, max_bytes: int) -> MaxBytesDeserializer[Self]:
        """Helper method to wrap the current deserializer with MaxBytesDeserializer."""
        from .adapters import MaxBytesDeserializer
        return MaxBytesDeserializer(self, max_bytes)

    @overload
    def with_optional_max_bytes(self, max_bytes: None) -> Self:
        ...

    @overload
    def with_optional_max_bytes(self, max_bytes: int) -> MaxBytesDeserializer[Self]:
        ...

    def with_optional_max_bytes(self, max_bytes: int | None) -> Self | MaxBytesDeserializer[Self]:
        """Helper method to optionally wrap the current deserializer."""
        if max_bytes is None:
            return self
        return self.with_max_bytes(max_bytes)



# Copyright 2025 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar

from typing_extensions import override

from hathorlib.serialization.deserializer import Deserializer
from hathorlib.serialization.exceptions import SerializationError
from hathorlib.serialization.serializer import Serializer

from ..types import Buffer
from .generic_adapter import GenericDeserializerAdapter, GenericSerializerAdapter

S = TypeVar('S', bound=Serializer)
D = TypeVar('D', bound=Deserializer)


class MaxBytesExceededError(SerializationError):
    """ This error is raised when the adapted serializer reached its maximum bytes write/read.

    After this exception is raised the adapted serializer cannot be used anymore. Handlers of this exception are
    expected to either: bubble up the exception (or an equivalente exception), or return an error. Handlers should not
    try to write again on the same serializer.

    It is possible that the inner serializer is still usable, but the point where the serialized stopped writing or
    reading might leave the rest of the data unusable, so for that reason it should be considered a failed
    (de)serialization overall, and not simply a failed "read/write" operation.
    """
    pass


class MaxBytesSerializer(GenericSerializerAdapter[S]):
    def __init__(self, serializer: S, max_bytes: int) -> None:
        super().__init__(serializer)
        self._bytes_left = max_bytes

    def _check_update_exceeds(self, write_size: int) -> None:
        self._bytes_left -= write_size
        if self._bytes_left < 0:
            raise MaxBytesExceededError

    @override
    def write_byte(self, data: int) -> None:
        self._check_update_exceeds(1)
        super().write_byte(data)

    @override
    def write_bytes(self, data: Buffer) -> None:
        data_view = memoryview(data)
        self._check_update_exceeds(len(data_view))
        super().write_bytes(data_view)


class MaxBytesDeserializer(GenericDeserializerAdapter[D]):
    def __init__(self, deserializer: D, max_bytes: int) -> None:
        super().__init__(deserializer)
        self._bytes_left = max_bytes

    def _check_update_exceeds(self, read_size: int) -> None:
        self._bytes_left -= read_size
        if self._bytes_left < 0:
            raise MaxBytesExceededError

    @override
    def read_byte(self) -> int:
        self._check_update_exceeds(1)
        return super().read_byte()

    @override
    def read_bytes(self, n: int, *, exact: bool = True) -> Buffer:
        self._check_update_exceeds(n)
        return super().read_bytes(n, exact=exact)

    @override
    def read_all(self) -> Buffer:
        result = super().read_bytes(self._bytes_left, exact=False)
        if not self.is_empty():
            raise MaxBytesExceededError
        return result



"""
Copyright (c) Hathor Labs and its affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""
import hashlib
import re
import struct
from typing import Any, Tuple, Union, cast

import base58
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from hathorlib.conf import HathorSettings
from hathorlib.exceptions import InvalidAddress
from hathorlib.serialization import Deserializer, SerializationError, Serializer
from hathorlib.serialization.adapters import MaxBytesExceededError
from hathorlib.serialization.encoding.leb128 import decode_leb128, encode_leb128

settings = HathorSettings()


def int_to_bytes(number: int, size: int, signed: bool = False) -> bytes:
    return number.to_bytes(size, byteorder='big', signed=signed)


def bytes_to_int(data: bytes, *, signed: bool = False) -> int:
    """
    Converts data in bytes to an int. Assumes big-endian format.

    Args:
        data: bytes to be converted
        signed: whether two's complement is used to represent the integer.

    Returns: the converted data as int
    """
    return int.from_bytes(data, byteorder='big', signed=signed)


def unpack(fmt: str, buf: bytes) -> Any:
    size = struct.calcsize(fmt)
    return struct.unpack(fmt, buf[:size]), buf[size:]


def unpack_len(n: int, buf: bytes) -> Tuple[bytes, bytes]:
    return buf[:n], buf[n:]


def get_checksum(address_bytes: bytes) -> bytes:
    """ Calculate double sha256 of address and gets first 4 bytes

        :param address_bytes: address before checksum
        :param address_bytes: bytes

        :return: checksum of the address
        :rtype: bytes
    """
    return hashlib.sha256(hashlib.sha256(address_bytes).digest()).digest()[:4]


def decode_address(address58: str) -> bytes:
    """ Decode address in base58 to bytes

    :param address58: Wallet address in base58
    :type address58: string

    :raises InvalidAddress: if address58 is not a valid base58 string or
                            not a valid address or has invalid checksum

    :return: Address in bytes
    :rtype: bytes
    """
    try:
        decoded_address = base58.b58decode(address58)
    except ValueError:
        # Invalid base58 string
        raise InvalidAddress('Invalid base58 address')
    # Validate address size [25 bytes]
    if len(decoded_address) != 25:
        raise InvalidAddress('Address size must have 25 bytes')
    # Validate the checksum
    address_checksum = decoded_address[-4:]
    valid_checksum = get_checksum(decoded_address[:-4])
    if address_checksum != valid_checksum:
        raise InvalidAddress('Invalid checksum of address')
    return decoded_address


def get_address_b58_from_public_key_hash(public_key_hash: bytes) -> str:
    """Gets the b58 address from the hash of a public key.

        :param public_key_hash: hash of public key (sha256 and ripemd160)
        :param public_key_hash: bytes

        :return: address in base 58
        :rtype: string
    """
    address = get_address_from_public_key_hash(public_key_hash)
    return base58.b58encode(address).decode('utf-8')


def get_address_from_public_key_hash(public_key_hash: bytes,
                                     version_byte: bytes = settings.P2PKH_VERSION_BYTE) -> bytes:
    """Gets the address in bytes from the public key hash

        :param public_key_hash: hash of public key (sha256 and ripemd160)
        :param public_key_hash: bytes

        :param version_byte: first byte of address to define the version of this address
        :param version_byte: bytes

        :return: address in bytes
        :rtype: bytes
    """
    address = b''
    # Version byte
    address += version_byte
    # Pubkey hash
    address += public_key_hash
    checksum = get_checksum(address)
    address += checksum
    return address


def get_address_b58_from_redeem_script_hash(redeem_script_hash: bytes,
                                            version_byte: bytes = settings.MULTISIG_VERSION_BYTE) -> str:
    """Gets the b58 address from the hash of the redeem script in multisig.

        :param redeem_script_hash: hash of the redeem script (sha256 and ripemd160)
        :param redeem_script_hash: bytes

        :return: address in base 58
        :rtype: string
    """
    address = get_address_from_redeem_script_hash(redeem_script_hash, version_byte)
    return base58.b58encode(address).decode('utf-8')


def get_address_from_redeem_script_hash(redeem_script_hash: bytes,
                                        version_byte: bytes = settings.MULTISIG_VERSION_BYTE) -> bytes:
    """Gets the address in bytes from the redeem script hash

        :param redeem_script_hash: hash of redeem script (sha256 and ripemd160)
        :param redeem_script_hash: bytes

        :param version_byte: first byte of address to define the version of this address
        :param version_byte: bytes

        :return: address in bytes
        :rtype: bytes
    """
    address = b''
    # Version byte
    address += version_byte
    # redeem script hash
    address += redeem_script_hash
    checksum = get_checksum(address)
    address += checksum
    return address


def clean_token_string(string: str) -> str:
    """ Receives the token name/symbol and returns it after some cleanups.
        It sets to uppercase, removes double spaces and spaces at the beginning and end.
    """
    return re.sub(r'\s\s+', ' ', string).strip().upper()


def get_public_key_from_bytes_compressed(public_key_bytes: bytes) -> ec.EllipticCurvePublicKey:
    """Return the cryptography public key from the compressed bytes format."""
    return ec.EllipticCurvePublicKey.from_encoded_point(ec.SECP256K1(), public_key_bytes)


def get_address_b58_from_public_key(public_key: ec.EllipticCurvePublicKey) -> str:
    """Get the b58 address from a public key."""
    public_key_bytes = get_public_key_bytes_compressed(public_key)
    return get_address_b58_from_public_key_bytes(public_key_bytes)


def get_address_b58_from_public_key_bytes(public_key_bytes: bytes) -> str:
    """Get the b58 address from a public key bytes."""
    public_key_hash = get_hash160(public_key_bytes)
    return get_address_b58_from_public_key_hash(public_key_hash)


def get_public_key_bytes_compressed(public_key: ec.EllipticCurvePublicKey) -> bytes:
    """Return the bytes of a pubkey in the compressed format."""
    return public_key.public_bytes(Encoding.X962, PublicFormat.CompressedPoint)


try:
    hashlib.new('ripemd160', b'')
except Exception:
    # XXX: the source says "Test-only pure Python RIPEMD160 implementation", however for our case this is acceptable
    #      for more details see: https://github.com/bitcoin/bitcoin/pull/23716/files which has a copy of the same code
    import pycoin.contrib.ripemd160  # type: ignore[import-untyped]

    def get_hash160(public_key_bytes: bytes) -> bytes:
        """The input is hashed twice: first with SHA-256 and then with RIPEMD-160"""
        key_hash = hashlib.sha256(public_key_bytes)
        return cast(bytes, pycoin.contrib.ripemd160.ripemd160(key_hash.digest()))
else:
    def get_hash160(public_key_bytes: bytes) -> bytes:
        """The input is hashed twice: first with SHA-256 and then with RIPEMD-160"""
        key_hash = hashlib.sha256(public_key_bytes)
        h = hashlib.new('ripemd160')
        h.update(key_hash.digest())
        return h.digest()


def encode_signed(value: int, *, max_bytes: Union[int, None] = None) -> bytes:
    """
    Receive a signed integer and return its LEB128-encoded bytes.

    >>> encode_signed(0) == bytes([0x00])
    True
    >>> encode_signed(624485) == bytes([0xE5, 0x8E, 0x26])
    True
    >>> encode_signed(-123456) == bytes([0xC0, 0xBB, 0x78])
    True
    """
    serializer: Serializer = Serializer.build_bytes_serializer()
    try:
        encode_leb128(serializer.with_optional_max_bytes(max_bytes), value, signed=True)
    except MaxBytesExceededError as e:
        raise ValueError(f'cannot encode more than {max_bytes} bytes') from e
    except SerializationError as e:
        raise ValueError('serialization error') from e
    return bytes(serializer.finalize())


def encode_unsigned(value: int, *, max_bytes: Union[int, None] = None) -> bytes:
    """
    Receive an unsigned integer and return its LEB128-encoded bytes.

    >>> encode_unsigned(0) == bytes([0x00])
    True
    >>> encode_unsigned(624485) == bytes([0xE5, 0x8E, 0x26])
    True
    """
    serializer: Serializer = Serializer.build_bytes_serializer()
    try:
        encode_leb128(serializer.with_optional_max_bytes(max_bytes), value, signed=False)
    except MaxBytesExceededError as e:
        raise ValueError(f'cannot encode more than {max_bytes} bytes') from e
    except SerializationError as e:
        raise ValueError('serialization error') from e
    return bytes(serializer.finalize())


def decode_signed(data: bytes, *, max_bytes: Union[int, None] = None) -> tuple[int, bytes]:
    """
    Receive and consume a buffer returning a tuple of the unpacked
    LEB128-encoded signed integer and the reamining buffer.

    >>> decode_signed(bytes([0x00]) + b'test')
    (0, b'test')
    >>> decode_signed(bytes([0xE5, 0x8E, 0x26]) + b'test')
    (624485, b'test')
    >>> decode_signed(bytes([0xC0, 0xBB, 0x78]) + b'test')
    (-123456, b'test')
    >>> decode_signed(bytes([0xC0, 0xBB, 0x78]) + b'test', max_bytes=3)
    (-123456, b'test')
    >>> try:
    ...     decode_signed(bytes([0xC0, 0xBB, 0x78]) + b'test', max_bytes=2)
    ... except ValueError as e:
    ...     print(e)
    cannot decode more than 2 bytes
    """
    deserializer = Deserializer.build_bytes_deserializer(data)
    try:
        value = decode_leb128(deserializer.with_optional_max_bytes(max_bytes), signed=True)
    except MaxBytesExceededError as e:
        raise ValueError(f'cannot decode more than {max_bytes} bytes') from e
    except SerializationError as e:
        raise ValueError('deserialization error') from e
    remaining_data = bytes(deserializer.read_all())
    deserializer.finalize()
    return (value, remaining_data)


def decode_unsigned(data: bytes, *, max_bytes: Union[int, None] = None) -> tuple[int, bytes]:
    """
    Receive and consume a buffer returning a tuple of the unpacked
    LEB128-encoded unsigned integer and the reamining buffer.

    >>> decode_unsigned(bytes([0x00]) + b'test')
    (0, b'test')
    >>> decode_unsigned(bytes([0xE5, 0x8E, 0x26]) + b'test')
    (624485, b'test')
    >>> decode_unsigned(bytes([0xE5, 0x8E, 0x26]) + b'test', max_bytes=3)
    (624485, b'test')
    >>> try:
    ...     decode_unsigned(bytes([0xE5, 0x8E, 0x26]) + b'test', max_bytes=2)
    ... except ValueError as e:
    ...     print(e)
    cannot decode more than 2 bytes
    """
    deserializer = Deserializer.build_bytes_deserializer(data)
    try:
        value = decode_leb128(deserializer.with_optional_max_bytes(max_bytes), signed=False)
    except MaxBytesExceededError as e:
        raise ValueError(f'cannot decode more than {max_bytes} bytes') from e
    except SerializationError as e:
        raise ValueError('deserialization error') from e
    remaining_data = bytes(deserializer.read_all())
    deserializer.finalize()
    return (value, remaining_data)



# Copyright 2025 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing_extensions import override

from .deserializer import Deserializer
from .exceptions import OutOfDataError
from .types import Buffer

_EMPTY_VIEW = memoryview(b'')


class BytesDeserializer(Deserializer):
    """Simple implementation of a Deserializer to parse values from a byte sequence.

    This implementation maintains a memoryview that is shortened as the bytes are read.
    """

    def __init__(self, data: Buffer) -> None:
        self._view = memoryview(data)

    @override
    def finalize(self) -> None:
        if not self.is_empty():
            raise ValueError('trailing data')
        del self._view

    @override
    def is_empty(self) -> bool:
        # XXX: least amount of OPs, "not" converts to bool with the correct semantics of "is empty"
        return not self._view

    @override
    def peek_byte(self) -> int:
        if not len(self._view):
            raise OutOfDataError('not enough bytes to read')
        return self._view[0]

    @override
    def peek_bytes(self, n: int, *, exact: bool = True) -> memoryview:
        if n < 0:
            raise ValueError('value cannot be negative')
        if exact and len(self._view) < n:
            raise OutOfDataError('not enough bytes to read')
        return self._view[:n]

    @override
    def read_byte(self) -> int:
        b = self.peek_byte()
        self._view = self._view[1:]
        return b

    @override
    def read_bytes(self, n: int, *, exact: bool = True) -> memoryview:
        b = self.peek_bytes(n, exact=exact)
        if exact and len(self._view) < n:
            raise OutOfDataError('not enough bytes to read')
        self._view = self._view[n:]
        return b

    @override
    def read_all(self) -> memoryview:
        b = self._view
        self._view = _EMPTY_VIEW
        return b



# Copyright 2024 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import NamedTuple

from hathorlib.conf import HathorSettings
from hathorlib.transaction import Transaction
from hathorlib.utils import int_to_bytes, unpack, unpack_len

settings = HathorSettings()

# used to allow new versions of the serialization format in the future
ON_CHAIN_BLUEPRINT_VERSION: int = 1


class CodeKind(Enum):
    """ Represents what type of code and format is being used, to allow new code/compression types in the future.
    """

    PYTHON_GZIP = 'python+gzip'

    def __bytes__(self) -> bytes:
        return self.value.encode()


class Code(NamedTuple):
    """ Store the code object in memory, along with helper methods.
    """

    # determines how the content will be interpreted
    kind: CodeKind

    # the encoded content, usually encoded implies compressed
    data: bytes

    def __bytes__(self) -> bytes:
        # Code serialization format: [kind:variable bytes][null byte][data:variable bytes]
        if self.kind is not CodeKind.PYTHON_GZIP:
            raise ValueError('Invalid code kind value')
        buf = bytearray()
        buf.extend(bytes(self.kind))
        buf.append(0)
        buf.extend(self.data)
        return bytes(buf)

    @classmethod
    def from_bytes(cls, data: bytes) -> 'Code':
        """ Parses a Code instance from a byte sequence, the length of the data is encoded outside of this class.

        NOTE: This will not validate whether the encoded has a valid compression format. A Validator must be used to
        check that.
        """
        data = bytearray(data)
        cut_at = data.index(0)
        kind = CodeKind(data[0:cut_at].decode())
        if kind is not CodeKind.PYTHON_GZIP:
            raise ValueError('Code kind not supported')
        compressed_code = data[cut_at + 1:]
        return cls(kind, compressed_code)


class OnChainBlueprint(Transaction):
    """On-chain blueprint vertex to be placed on the DAG of transactions."""

    MIN_NUM_INPUTS = 0

    def __init__(self) -> None:
        super().__init__()

        # Pubkey and signature of the transaction owner / caller.
        self.nc_pubkey: bytes = b''
        self.nc_signature: bytes = b''

        self.code: Code = Code(CodeKind.PYTHON_GZIP, b'')

    def serialize_code(self) -> bytes:
        """Serialization of self.code, to be used for the serialization of this transaction type."""
        buf = bytearray()
        buf.extend(int_to_bytes(ON_CHAIN_BLUEPRINT_VERSION, 1))
        serialized_code = bytes(self.code)
        buf.extend(int_to_bytes(len(serialized_code), 4))
        buf.extend(serialized_code)
        return bytes(buf)

    @classmethod
    def deserialize_code(_cls, buf: bytes) -> tuple[Code, bytes]:
        """Parses the self.code field, returns the parse result and the remaining bytes."""
        (ocb_version,), buf = unpack('!B', buf)
        if ocb_version != ON_CHAIN_BLUEPRINT_VERSION:
            raise ValueError(f'unknown on-chain blueprint version: {ocb_version}')

        (serialized_code_len,), buf = unpack('!L', buf)
        max_serialized_code_len = settings.NC_ON_CHAIN_BLUEPRINT_CODE_MAX_SIZE_COMPRESSED
        if serialized_code_len > max_serialized_code_len:
            raise ValueError(f'compressed code data is too large: {serialized_code_len} > {max_serialized_code_len}')
        serialized_code, buf = unpack_len(serialized_code_len, buf)
        code = Code.from_bytes(serialized_code)
        return code, buf

    def _serialize_ocb(self, *, skip_signature: bool = False) -> bytes:
        buf = bytearray()
        buf += self.serialize_code()
        buf += int_to_bytes(len(self.nc_pubkey), 1)
        buf += self.nc_pubkey
        if not skip_signature:
            buf += int_to_bytes(len(self.nc_signature), 1)
            buf += self.nc_signature
        else:
            buf += int_to_bytes(0, 1)
        return bytes(buf)

    def get_funds_struct(self) -> bytes:
        struct_bytes = super().get_funds_struct()
        struct_bytes += self._serialize_ocb()
        return struct_bytes

    def get_funds_fields_from_struct(self, buf: bytes) -> bytes:
        buf = super().get_funds_fields_from_struct(buf)

        code, buf = OnChainBlueprint.deserialize_code(buf)
        self.code = code

        (nc_pubkey_len,), buf = unpack('!B', buf)
        self.nc_pubkey, buf = unpack_len(nc_pubkey_len, buf)
        (nc_signature_len,), buf = unpack('!B', buf)
        self.nc_signature, buf = unpack_len(nc_signature_len, buf)

        return buf



from .generic_adapter import GenericDeserializerAdapter, GenericSerializerAdapter
from .max_bytes import MaxBytesDeserializer, MaxBytesExceededError, MaxBytesSerializer

__all__ = [
    'GenericDeserializerAdapter',
    'GenericSerializerAdapter',
    'MaxBytesDeserializer',
    'MaxBytesExceededError',
    'MaxBytesSerializer',
]



# Copyright (c) Hathor Labs and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
import json
import re
from typing import Any, Dict, List, NamedTuple, Optional, cast
from urllib.parse import urljoin

from hathorlib.base_transaction import tx_or_block_from_bytes
from hathorlib.exceptions import PushTxFailed

try:
    from aiohttp import ClientSession
    from structlog import get_logger
except ImportError as e:
    raise ImportError('Missing dependency, please install extras: hathorlib[client]') from e

from hathorlib import Block, TxOutput

REQUIRED_HATHOR_API_VERSION = 'v1a'

logger = get_logger()


# This regex was copied from https://semver.org/#is-there-a-suggested-regular-expression-regex-to-check-a-semver-string
semver_pattern = (
    r"(?P<major>0|[1-9]\d*)\.(?P<minor>0|[1-9]\d*)\.(?P<patch>0|[1-9]\d*)"
    r"(?:-(?P<prerelease>(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*)(?:\.(?:0|[1-9]\d*|\d*[a-zA-Z-][0-9a-zA-Z-]*))*))?"
    r"(?:\+(?P<metadata>[0-9a-zA-Z-]+(?:\.[0-9a-zA-Z-]+)*))?"
)
semver_re = re.compile(semver_pattern)


class BlockTemplate(NamedTuple):
    """Block template."""

    data: bytes
    height: int

    def to_dict(self) -> Dict[str, Any]:
        """Return dict for json serialization."""
        return {
            'data': self.data.hex(),
            'height': self.height,
        }


class HathorVersion(NamedTuple):
    """Hathor backend version."""

    major: int
    minor: int
    patch: int
    prerelease: Optional[str] = None
    metadata: Optional[str] = None


class HathorClient:
    """Used to communicate with Hathor's full-node."""

    USER_AGENT = 'tx-mining-service'

    def __init__(self, server_url: str, api_version: str = REQUIRED_HATHOR_API_VERSION):
        """Init HathorClient with a Hathor's full-node backend."""
        self.log = logger.new()
        self._base_url = urljoin(server_url, api_version).rstrip('/') + '/'
        self._base_headers = {
            'User-Agent': self.USER_AGENT,
        }
        self._session: Optional[ClientSession] = None

    async def start(self) -> None:
        """Start a session with the backend."""
        self._session = ClientSession(headers=self._base_headers)

    async def stop(self) -> None:
        """Stop a session with the backend."""
        if self._session is not None:
            await self._session.close()
            self._session = None

    def _get_url(self, url: str) -> str:
        return urljoin(self._base_url, url.lstrip('/'))

    async def version(self) -> HathorVersion:
        """Return the version of the backend."""
        assert self._session is not None

        async with self._session.get(self._get_url('version')) as resp:
            data = await resp.json()
            version = data['version']

            match = semver_re.match(version)

            if match:
                result = match.groupdict()

                return HathorVersion(
                    major=int(result['major']),
                    minor=int(result['minor']),
                    patch=int(result['patch']),
                    prerelease=result.get('prerelease'),
                    metadata=result.get('metadata'),
                )
            else:
                raise RuntimeError(f'Cannot parse version {version}')

    async def health(self) -> Dict[str, Any]:
        """Return the health information of the backend."""
        assert self._session is not None

        async with self._session.get(self._get_url('health')) as resp:
            data = await resp.text()
            try:
                parsed_json: Dict[str, Any] = json.loads(data)
            except json.JSONDecodeError:
                raise RuntimeError('Cannot parse health response: {}'.format(data))
            return parsed_json

    async def get_block_template(self, address: Optional[str] = None) -> BlockTemplate:
        """Return a block template."""
        assert self._session is not None
        params = {}
        if address is not None:
            params['address'] = address
        resp = await self._session.get(self._get_url('get_block_template'), params=params)
        if resp.status != 200:
            self.log.error('Error getting block template', status=resp.status)
            raise RuntimeError('Cannot get block template (status {})'.format(resp.status))

        data = await resp.json()

        if data.get('error'):
            self.log.error('Error getting block template', data=data)
            raise RuntimeError('Cannot get block template')

        # Get height.
        metadata = data.get('metadata', {})
        height = metadata['height']

        # Build block.
        blk = Block()
        blk.signal_bits = data['signal_bits']
        blk.version = 0
        blk.timestamp = data['timestamp']
        blk.weight = data['weight']
        blk.parents = [bytes.fromhex(x) for x in data['parents']]
        blk.data = b''

        do = data['outputs'][0]
        txout = TxOutput(
            value=do['value'],
            token_data=0,
            script=b'',
        )
        blk.outputs = [txout]
        return BlockTemplate(data=bytes(blk), height=height)

    async def get_tx_parents(self) -> List[bytes]:
        """Return parents for a new transaction."""
        assert self._session is not None
        async with self._session.get(self._get_url('tx_parents')) as resp:
            data = await resp.json()
            if not data.get('success'):
                raise RuntimeError('Cannot get tx parents')
            return [bytes.fromhex(x) for x in data['tx_parents']]

    async def push_tx_or_block(self, raw: bytes) -> bool:
        """Push a new tx or block to the backend."""
        assert self._session is not None

        tx = tx_or_block_from_bytes(raw)

        if tx.is_block:
            data = {'hexdata': raw.hex()}
            resp = await self._session.post(self._get_url('submit_block'), json=data)
        else:
            data = {'hex_tx': raw.hex()}
            resp = await self._session.post(self._get_url('push_tx'), json=data)

        status = resp.status
        if status > 299:
            response = await resp.text()
            self.log.error('Error pushing tx or block', response=response, status=status)
            raise PushTxFailed('Cannot push tx or block')

        json = await resp.json()

        return cast(bool, json['result'])



# Copyright 2023 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from hathorlib.base_transaction import BaseTransaction


class VertexBaseHeader(ABC):
    @classmethod
    @abstractmethod
    def deserialize(cls, tx: BaseTransaction, buf: bytes) -> tuple[VertexBaseHeader, bytes]:
        """Deserialize header from `buf` which starts with header id."""
        raise NotImplementedError

    @abstractmethod
    def serialize(self) -> bytes:
        """Serialize header with header id as prefix."""
        raise NotImplementedError

    @abstractmethod
    def get_sighash_bytes(self) -> bytes:
        """Return sighash bytes to check digital signatures."""
        raise NotImplementedError



"""
Copyright 2019 Hathor Labs

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class HathorError(Exception):
    """General error class"""


class InvalidAddress(HathorError):
    """Address is invalid"""


class TxValidationError(HathorError):
    """Base class for tx validation errors"""


class HathorClientError(HathorError):
    """Base class for errors when communicating with the fullnode"""


class ParentDoesNotExist(TxValidationError):
    """A parent does not exist"""


class IncorrectParents(TxValidationError):
    """Wrong number of parents or confirming incorrect types of transactions:
    - block: 3 parents: 1 block, 2 transactions
    - tx: 2 parents, both transactions
    """


class TimestampError(TxValidationError):
    """Transaction timestamp is smaller or equal to one parent's timestamp"""


class DoubleSpend(TxValidationError):
    """Some input has already been spent"""


class InputOutputMismatch(TxValidationError):
    """Input and output amounts are not equal"""


class InvalidInputData(TxValidationError):
    """Input data does not solve output script correctly"""


class NoInputError(TxValidationError):
    """There is not input"""


class TooManyInputs(TxValidationError):
    """More than 256 inputs"""


class InexistentInput(TxValidationError):
    """Input tx does not exist or index spent does not exist"""


class ConflictingInputs(TxValidationError):
    """Inputs in the tx are spending the same output"""


class TooManyOutputs(TxValidationError):
    """More than 256 outputs"""


class InvalidOutputValue(TxValidationError):
    """Value of output is invalid"""


class PowError(TxValidationError):
    """Proof-of-work is not correct"""


class AuxPowError(PowError):
    """Auxiliary Proof-of-work is not correct"""


class WeightError(TxValidationError):
    """Transaction not using correct weight"""


class DuplicatedParents(TxValidationError):
    """Transaction has duplicated parents"""


class InvalidToken(TxValidationError):
    """Token is not valid"""


class BlockError(TxValidationError):
    """Base class for Block-specific errors"""


class TransactionDataError(TxValidationError):
    """Block data max length exceeded"""


class RewardLocked(TxValidationError):
    """Block reward cannot be spent yet, needs more confirmations"""


class BlockWithInputs(BlockError):
    """Block has inputs"""


class BlockWithTokensError(BlockError):
    """Block has tokens other than hathor"""


class ScriptError(HathorError):
    """Base class for script evaluation errors"""


class OutOfData(ScriptError):
    """PUSHDATA operation with more bytes than we have available"""


class MissingStackItems(ScriptError):
    """Operation requires more items than what is on stack"""


class EqualVerifyFailed(ScriptError):
    """OP_EQUALVERIFY failed"""


class FinalStackInvalid(ScriptError):
    """Value left on stack is not true"""


class OracleChecksigFailed(ScriptError):
    """Signature, public key and data don't match. Used mostly with nano contracts"""


class DataIndexError(ScriptError):
    """The value for data at the given index does not exist.

    For example, if the data is of form 'value1:value2:value3' and we try to access value at index 5.
    """


class InvalidStackData(ScriptError):
    """The value for data on the stack is not what we expect

    For example, we expect an integer but it's not
    """


class VerifyFailed(ScriptError):
    """For all cases when there's a comparison that fails"""


class TimeLocked(ScriptError):
    """Transaction is invalid because it is time locked"""


class PushTxFailed(HathorClientError):
    """An attempt to push a tx/block to the fullnode failed"""



# Copyright (c) Hathor Labs and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from math import log
from typing import TYPE_CHECKING

from hathorlib.conf import HathorSettings

if TYPE_CHECKING:
    from hathorlib import Transaction

settings = HathorSettings()


def minimum_tx_weight(tx: 'Transaction', *, fix_parents: bool = True) -> float:
    """Return the minimum weight for the tx.

       The minimum is calculated by the following function:

       w = alpha * log(size, 2) +       4.0         + 4.0
                                  ----------------
                                   1 + k / amount
    """
    tx_size = len(tx.get_struct())

    # When a transaction is still being create, it might not have its parents yet.
    # In this case, the parents will be added later but we need to take their size
    # into consideration to calculate the weight.
    if fix_parents and len(tx.parents) < 2:
        tx_size += 32 * (2 - len(tx.parents))

    # We need to take into consideration the decimal places because it is inside the amount.
    # For instance, if one wants to transfer 20 HTRs, the amount will be 2000.
    # Max below is preventing division by 0 when handling authority methods that have no outputs
    amount = max(1, tx.sum_outputs) / (10 ** settings.DECIMAL_PLACES)

    weight: float = (
        + settings.MIN_TX_WEIGHT_COEFFICIENT * log(tx_size, 2)
        + 4 / (1 + settings.MIN_TX_WEIGHT_K / amount) + 4
    )

    # Make sure the calculated weight is at least the minimum
    weight = max(weight, settings.MIN_TX_WEIGHT)

    return weight



# Copyright 2025 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import struct
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, TypeVar, overload

from typing_extensions import Self

from .types import Buffer

if TYPE_CHECKING:
    from .adapters import MaxBytesSerializer
    from .bytes_serializer import BytesSerializer

T = TypeVar('T')


class Serializer(ABC):
    def finalize(self) -> Buffer:
        """Get the resulting byte sequence, the serializer cannot be reused after this."""
        raise TypeError('this serializer does not support finalization')

    @abstractmethod
    def cur_pos(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def write_byte(self, data: int) -> None:
        """Write a single byte."""
        raise NotImplementedError

    @abstractmethod
    def write_bytes(self, data: Buffer) -> None:
        # XXX: it is recommended that implementors of Serializer specialize this implementation
        for byte in bytes(memoryview(data)):
            self.write_byte(byte)

    def write_struct(self, data: tuple[Any, ...], format: str) -> None:
        data_bytes = struct.pack(format, *data)
        self.write_bytes(data_bytes)

    def with_max_bytes(self, max_bytes: int) -> MaxBytesSerializer[Self]:
        """Helper method to wrap the current serializer with MaxBytesSerializer."""
        from .adapters import MaxBytesSerializer
        return MaxBytesSerializer(self, max_bytes)

    @overload
    def with_optional_max_bytes(self, max_bytes: None) -> Self:
        ...

    @overload
    def with_optional_max_bytes(self, max_bytes: int) -> MaxBytesSerializer[Self]:
        ...

    def with_optional_max_bytes(self, max_bytes: int | None) -> Self | MaxBytesSerializer[Self]:
        """Helper method to optionally wrap the current serializer."""
        if max_bytes is None:
            return self
        return self.with_max_bytes(max_bytes)

    @staticmethod
    def build_bytes_serializer() -> BytesSerializer:
        from .bytes_serializer import BytesSerializer
        return BytesSerializer()



# Copyright 2023 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hathorlib.headers.base import VertexBaseHeader
from hathorlib.headers.deprecated_nano_header import DeprecatedNanoHeader
from hathorlib.headers.nano_header import NC_INITIALIZE_METHOD, NanoHeader
from hathorlib.headers.types import VertexHeaderId

__all__ = [
    'VertexBaseHeader',
    'VertexHeaderId',
    'NanoHeader',
    'DeprecatedNanoHeader',
    'NC_INITIALIZE_METHOD',
]



"""
Copyright (c) Hathor Labs and its affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from struct import pack
from typing import Dict

from hathorlib.base_transaction import BaseTransaction, TxOutput
from hathorlib.utils import int_to_bytes, unpack, unpack_len

# Signal bits (B), version (B), outputs len (B)
_FUNDS_FORMAT_STRING = '!BBB'

# Signal bits (B), version (B), inputs len (B) and outputs len (B)
_SIGHASH_ALL_FORMAT_STRING = '!BBBB'


class Block(BaseTransaction):
    SERIALIZATION_NONCE_SIZE = 16

    @property
    def is_block(self) -> bool:
        """Returns true if this is a block"""
        return True

    @property
    def is_transaction(self) -> bool:
        """Returns true if this is a transaction"""
        return False

    def _get_formatted_fields_dict(self, short: bool = True) -> Dict[str, str]:
        d = super()._get_formatted_fields_dict(short)
        if not short:
            d.update(data=self.data.hex())
        return d

    @classmethod
    def create_from_struct(cls, struct_bytes: bytes) -> 'Block':
        blc = cls()
        buf = blc.get_fields_from_struct(struct_bytes)

        if len(buf) < cls.SERIALIZATION_NONCE_SIZE:
            raise ValueError('Invalid sequence of bytes')

        blc.nonce = int.from_bytes(buf[:cls.SERIALIZATION_NONCE_SIZE], byteorder='big')
        buf = buf[cls.SERIALIZATION_NONCE_SIZE:]

        while buf:
            buf = blc.get_header_from_bytes(buf)

        blc.hash = blc.calculate_hash()

        return blc

    def get_funds_fields_from_struct(self, buf: bytes) -> bytes:
        """ Gets all funds fields for a block from a buffer.

        :param buf: Bytes of a serialized block
        :type buf: bytes

        :return: A buffer containing the remaining struct bytes
        :rtype: bytes

        :raises ValueError: when the sequence of bytes is incorect
        """
        (self.signal_bits, self.version, outputs_len), buf = unpack(_FUNDS_FORMAT_STRING, buf)

        for _ in range(outputs_len):
            txout, buf = TxOutput.create_from_bytes(buf)
            self.outputs.append(txout)

        return buf

    def get_graph_fields_from_struct(self, buf: bytes) -> bytes:
        """ Gets graph fields for a block from a buffer.

        :param buf: Bytes of a serialized transaction
        :type buf: bytes

        :return: A buffer containing the remaining struct bytes
        :rtype: bytes

        :raises ValueError: when the sequence of bytes is incorect
        """
        buf = super().get_graph_fields_from_struct(buf)
        (data_bytes,), buf = unpack('!B', buf)
        self.data, buf = unpack_len(data_bytes, buf)
        return buf

    def get_funds_struct(self) -> bytes:
        """Return the funds data serialization of the block

        :return: funds data serialization of the block
        :rtype: bytes
        """
        struct_bytes = pack(_FUNDS_FORMAT_STRING, self.signal_bits, self.version, len(self.outputs))

        for tx_output in self.outputs:
            struct_bytes += bytes(tx_output)

        return struct_bytes

    def get_graph_struct(self) -> bytes:
        """Return the graph data serialization of the block, without including the nonce field

        :return: graph data serialization of the transaction
        :rtype: bytes
        """
        struct_bytes_without_data = super().get_graph_struct()
        data_bytes = int_to_bytes(len(self.data), 1)
        return struct_bytes_without_data + data_bytes + self.data



# Copyright 2023 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections import deque
from dataclasses import dataclass
from typing import TYPE_CHECKING

from hathorlib.headers.base import VertexBaseHeader
from hathorlib.headers.types import VertexHeaderId
from hathorlib.utils import int_to_bytes, unpack, unpack_len

if TYPE_CHECKING:
    from hathorlib.base_transaction import BaseTransaction
    from hathorlib.headers.nano_header import NanoHeader, NanoHeaderAction


NC_VERSION = 1


@dataclass(frozen=True)
class DeprecatedNanoHeader(VertexBaseHeader):
    tx: BaseTransaction

    # nc_id equals to the blueprint_id when a Nano Contract is being created.
    # nc_id equals to the nanocontract_id when a method is being called.
    nc_id: bytes

    # Name of the method to be called. When creating a new Nano Contract, it must be equal to 'initialize'.
    nc_method: str

    # Serialized arguments to nc_method.
    nc_args_bytes: bytes

    nc_actions: list[NanoHeaderAction]

    # Pubkey and signature of the transaction owner / caller.
    nc_pubkey: bytes
    nc_signature: bytes

    nc_version: int = NC_VERSION

    @classmethod
    def deserialize(cls, tx: BaseTransaction, buf: bytes) -> tuple[DeprecatedNanoHeader, bytes]:
        header_id, buf = buf[:1], buf[1:]
        assert header_id == VertexHeaderId.NANO_HEADER.value
        (nc_version,), buf = unpack('!B', buf)
        if nc_version != NC_VERSION:
            raise ValueError('unknown nanocontract version: {}'.format(nc_version))

        nc_id, buf = unpack_len(32, buf)
        (nc_method_len,), buf = unpack('!B', buf)
        nc_method, buf = unpack_len(nc_method_len, buf)
        (nc_args_bytes_len,), buf = unpack('!H', buf)
        nc_args_bytes, buf = unpack_len(nc_args_bytes_len, buf)

        nc_actions: list[NanoHeaderAction] = []
        from hathorlib.nanocontracts import DeprecatedNanoContract
        if not isinstance(tx, DeprecatedNanoContract):
            (nc_actions_len,), buf = unpack('!B', buf)
            for _ in range(nc_actions_len):
                action, buf = NanoHeader._deserialize_action(buf)
                nc_actions.append(action)

        (nc_pubkey_len,), buf = unpack('!B', buf)
        nc_pubkey, buf = unpack_len(nc_pubkey_len, buf)
        (nc_signature_len,), buf = unpack('!B', buf)
        nc_signature, buf = unpack_len(nc_signature_len, buf)

        decoded_nc_method = nc_method.decode('ascii')

        return cls(
            tx=tx,
            nc_version=nc_version,
            nc_id=nc_id,
            nc_method=decoded_nc_method,
            nc_args_bytes=nc_args_bytes,
            nc_actions=nc_actions,
            nc_pubkey=nc_pubkey,
            nc_signature=nc_signature,
        ), bytes(buf)

    def _serialize_without_header_id(self, *, skip_signature: bool) -> deque[bytes]:
        """Serialize the header with the option to skip the signature."""
        encoded_method = self.nc_method.encode('ascii')

        ret: deque[bytes] = deque()
        ret.append(int_to_bytes(NC_VERSION, 1))
        ret.append(self.nc_id)
        ret.append(int_to_bytes(len(encoded_method), 1))
        ret.append(encoded_method)
        ret.append(int_to_bytes(len(self.nc_args_bytes), 2))
        ret.append(self.nc_args_bytes)

        from hathorlib.nanocontracts import DeprecatedNanoContract
        if not isinstance(self.tx, DeprecatedNanoContract):
            ret.append(int_to_bytes(len(self.nc_actions), 1))
            for action in self.nc_actions:
                ret.append(NanoHeader._serialize_action(action))

        ret.append(int_to_bytes(len(self.nc_pubkey), 1))
        ret.append(self.nc_pubkey)
        if not skip_signature:
            ret.append(int_to_bytes(len(self.nc_signature), 1))
            ret.append(self.nc_signature)
        else:
            ret.append(int_to_bytes(0, 1))
        return ret

    def serialize(self) -> bytes:
        ret = self._serialize_without_header_id(skip_signature=False)
        ret.appendleft(VertexHeaderId.NANO_HEADER.value)
        return b''.join(ret)

    def get_sighash_bytes(self) -> bytes:
        ret = self._serialize_without_header_id(skip_signature=True)
        return b''.join(ret)



# Copyright 2025 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing_extensions import override

from .serializer import Serializer
from .types import Buffer


class BytesSerializer(Serializer):
    """Simple implementation of Serializer to write to memory.

    This implementation defers joining everything until finalize is called, before that every write is stored as a
    memoryview in a list.
    """

    def __init__(self) -> None:
        self._parts: list[memoryview] = []
        self._pos: int = 0

    @override
    def finalize(self) -> memoryview:
        result = memoryview(b''.join(self._parts))
        del self._parts
        del self._pos
        return result

    @override
    def cur_pos(self) -> int:
        return self._pos

    @override
    def write_byte(self, data: int) -> None:
        # int.to_bytes checks for correct range
        self._parts.append(memoryview(int.to_bytes(data, length=1, byteorder='big')))
        self._pos += 1

    @override
    def write_bytes(self, data: Buffer) -> None:
        part = memoryview(data)
        self._parts.append(part)
        self._pos += len(part)



# Copyright 2025 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import TracebackType
from typing import Generic, TypeVar, Union

from typing_extensions import Self, override

from hathorlib.serialization.deserializer import Deserializer
from hathorlib.serialization.serializer import Serializer

from ..types import Buffer

S = TypeVar('S', bound=Serializer)
D = TypeVar('D', bound=Deserializer)


class GenericSerializerAdapter(Serializer, Generic[S]):
    inner: S

    def __init__(self, serializer: S) -> None:
        self.inner = serializer

    @override
    def finalize(self) -> Buffer:
        return self.inner.finalize()

    @override
    def cur_pos(self) -> int:
        return self.inner.cur_pos()

    @override
    def write_byte(self, data: int) -> None:
        self.inner.write_byte(data)

    @override
    def write_bytes(self, data: Buffer) -> None:
        self.inner.write_bytes(data)

    # allow using this adapter as a context manager:

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Union[type[BaseException], None],
        exc_value: Union[BaseException, None],
        traceback: Union[TracebackType, None],
    ) -> None:
        pass


class GenericDeserializerAdapter(Deserializer, Generic[D]):
    inner: D

    def __init__(self, deserializer: D) -> None:
        self.inner = deserializer

    @override
    def finalize(self) -> None:
        return self.inner.finalize()

    @override
    def is_empty(self) -> bool:
        return self.inner.is_empty()

    @override
    def peek_byte(self) -> int:
        return self.inner.peek_byte()

    @override
    def peek_bytes(self, n: int, *, exact: bool = True) -> Buffer:
        return self.inner.peek_bytes(n, exact=exact)

    @override
    def read_byte(self) -> int:
        return self.inner.read_byte()

    @override
    def read_bytes(self, n: int, *, exact: bool = True) -> Buffer:
        return self.inner.read_bytes(n, exact=exact)

    @override
    def read_all(self) -> Buffer:
        return self.inner.read_all()

    # allow using this adapter as a context manager:

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Union[type[BaseException], None],
        exc_value: Union[BaseException, None],
        traceback: Union[TracebackType, None],
    ) -> None:
        pass



# Copyright 2023 Hathor Labs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from hathorlib import Transaction, TxVersion


class DeprecatedNanoContract(Transaction):
    """NanoContract vertex to be placed on the DAG of transactions."""

    def __init__(self) -> None:
        super().__init__()

        self.version = TxVersion.NANO_CONTRACT

        # nc_id equals to the blueprint_id when a Nano Contract is being created.
        # nc_id equals to the nanocontract_id when a method is being called.
        self.nc_id: bytes = b''

        # Name of the method to be called. When creating a new Nano Contract, it must be equal to 'initialize'.
        self.nc_method: str = ''

        # Serialized arguments to nc_method.
        self.nc_args_bytes: bytes = b''

        # Pubkey and signature of the transaction owner / caller.
        self.nc_pubkey: bytes = b''
        self.nc_signature: bytes = b''

    ################################
    # Methods for Transaction
    ################################

    def get_funds_fields_from_struct(self, buf: bytes) -> bytes:
        from hathorlib.headers import DeprecatedNanoHeader, VertexHeaderId
        buf = super().get_funds_fields_from_struct(buf)
        nano_header, buf = DeprecatedNanoHeader.deserialize(self, VertexHeaderId.NANO_HEADER.value + buf)
        self.headers.append(nano_header)
        return buf

    def get_funds_struct(self) -> bytes:
        struct_bytes = super().get_funds_struct()
        nano_header_bytes = self.get_nano_header().serialize()
        struct_bytes += nano_header_bytes[1:]
        return struct_bytes

    def get_headers_hash(self) -> bytes:
        return b''

    def get_headers_struct(self) -> bytes:
        return b''




from hathorlib.base_transaction import BaseTransaction, TxInput, TxOutput, TxVersion, sum_weights
from hathorlib.block import Block
from hathorlib.token_creation_tx import TokenCreationTransaction
from hathorlib.transaction import Transaction

__all__ = [
    'BaseTransaction',
    'Block',
    'TokenCreationTransaction',
    'Transaction',
    'TxInput',
    'TxOutput',
    'TxVersion',
    'sum_weights',
]



#  Copyright 2025 Hathor Labs
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
This module implements LEB128 for signed integers.

LEB128 or Little Endian Base 128 is a variable-length code compression used to store arbitrarily large
integers in a small number of bytes. LEB128 is used in the DWARF debug file format and the WebAssembly
binary encoding for all integer literals.

References:
- https://en.wikipedia.org/wiki/LEB128
- https://dwarfstd.org/doc/DWARF5.pdf
- https://webassembly.github.io/spec/core/binary/values.html#integers

This module implements LEB128 encoding/decoding using the standard 1-byte block split into 1-bit for continuation and
7-bits for data. The data can be either a signed or unsigned integer.

>>> se = Serializer.build_bytes_serializer()
>>> se.write_bytes(b'test')  # writes 74657374
>>> encode_leb128(se, 0, signed=True)  # writes 00
>>> encode_leb128(se, 624485, signed=True)  # writes e58e26
>>> encode_leb128(se, -123456, signed=True)  # writes c0bb78
>>> bytes(se.finalize()).hex()
'7465737400e58e26c0bb78'

>>> data = bytes.fromhex('00 e58e26 c0bb78 74657374')
>>> de = Deserializer.build_bytes_deserializer(data)
>>> decode_leb128(de, signed=True)  # reads 00
0
>>> decode_leb128(de, signed=True)  # reads e58e26
624485
>>> decode_leb128(de, signed=True)  # reads c0bb78
-123456
>>> bytes(de.read_all())  # reads 74657374
b'test'
>>> de.finalize()
"""

from hathorlib.serialization import Deserializer, Serializer


def encode_leb128(serializer: Serializer, value: int, *, signed: bool) -> None:
    """ Encodes an integer using LEB128.

    Caller must explicitly choose `signed=True` or `signed=False`.

    This module's docstring has more details on LEB128 and examples.
    """
    if not signed and value < 0:
        raise ValueError('cannot encode value <0 as unsigend')
    while True:
        byte = value & 0b0111_1111
        value >>= 7
        if signed:
            cont = (value == 0 and (byte & 0b0100_0000) == 0) or (value == -1 and (byte & 0b0100_0000) != 0)
        else:
            cont = (value == 0 and (byte & 0b1000_0000) == 0)
        if cont:
            serializer.write_byte(byte)
            break
        serializer.write_byte(byte | 0b1000_0000)


def decode_leb128(deserializer: Deserializer, *, signed: bool) -> int:
    """ Decodes a LEB128-encoded integer.

    Caller must explicitly choose `signed=True` or `signed=False`.

    This module's docstring has more details on LEB128 and examples.
    """
    result = 0
    shift = 0
    while True:
        byte = deserializer.read_byte()
        result |= (byte & 0b0111_1111) << shift
        shift += 7
        assert shift % 7 == 0
        if (byte & 0b1000_0000) == 0:
            if signed and (byte & 0b0100_0000) != 0:
                return result | -(1 << shift)
            return result
